#!/usr/bin/perl

use strict;
use warnings;

# It detects occurrences of Simpson's paradox in a dataset, 
# as reported in the following published article:
# Briefings in Bioinformatics, Published online (ahead of
# print) on 9 Jan. 2019, DOI:10.1093/bib/bby126
# https://academic.oup.com/bib/advance
# -article/doi/10.1093/bib/bby126/5280899

# Input: file in the ARFF format used by WEKA data mining tool
# Output: two files: Simpsons-Paradoxes.txt, and
#                    LOG-Simpsons-Paradoxes.txt

#############################################################
# The goal attribute id, goal attr. value of interest id,   #
# goal attr. value of interest name and the attribute ids   #
# in the list of predictor (first-partitioning) attributes  #
# and the list of confounder (second-partitioning) attrib's #
# are HARD-CODED in the program.                            #
# So the following lines MUST BE UPDATED for each dataset:  #                      
# Lines setting variables: $GA_ID, $GOAL_VAL, $GOAL_VAL_NAME#
# @part_attr_1_list (list of candidate predictor attributes)#
# @part_attr_2_list (list of candidate confounder attrib's) #
# Note that attribute ids start to count from 0, not 1      #
# Hence, attr id in perl = attr id in WEKA result - 1       #
#############################################################
# The below hard-coded values are associated with the       #
# input file: Mouse-GO-attrs-class-anti-longev.arff.txt     #
#############################################################
my $GA_ID = 886; # Goal (Class) Attr. Id in Mouse dataset 
my $GOAL_VAL = '0'; # Goal Value as recorded in input file
my $GOAL_VAL_NAME = 'anti-longevity'; # meaningful value name
### Top-15 predictor attributes in ranking by InfoGain    ###
my @part_attr_1_list = (3,737,736,790,805,732,655,54,72,89,101,619,403,401,71);
### The list of confounders has all 886 predictor attribs ### 
my @part_attr_2_list = (0..885);
########### END OF VARIABLES WITH HARD-CODED VALUES #########

# Warning:
# It checks if the goal, predictor and confounder attributes 
# are nominal; but it does not check if the goal and predictor
# attributes are binary, which are required conditions for 
# the program to work correctly. So, users have to ensure 
# these conditions are true.

# Terminology issues:
# The terms 1st partitioning attribute, 2nd partitioning
# attribute and goal attribute used in this program correspond
# to the terms predictor, confounder and class attributes
# (variables) in the aforementioned published article.

my $input_file = $ARGV[0];
open INPUT, "$input_file"
  or die "Cannot open input file: $!";
open OUTPUT, ">Simpsons-Paradoxes.txt" 
  or die "Cannot open output file: $!";
open LOG, ">LOG-Simpsons-Paradoxes.txt"
  or die "Cannot open output file: $!";
print OUTPUT "INPUT FILE: $input_file\n";

my @attr_names = (); # 1D array:[attr_id]
my $attr_name = '';
my @attr_types = (); # 1D array:[attr_id]
my @attr_values = (); # 2D array:[attr_id][value]
my $attr_value = '';
my $attr_values_or_type = ''; # used to read attribute line
my $attr_count = 0; # attr. counter when reading input file  
my $ATTR_COUNT = 0; # No. of predictor + class attrs; class id
my $expected_line_type = '';
my @dataset = ();   # 2D array:[inst][attr_id]
my @instance = ();  # 1D array:[attr_id] 
my $inst_count = 0; # instance counter when reading input file
my $INST_COUNT = 0; # Number of instances read from input file 

my $pa1_id = ''; # attr. id for loop over @part_attr_1_list
my $pa1_v = ''; # partitioning attr. 1's current value 
my $pa1_v1 = ''; # partitioning attr. 1's value 1 
my $pa1_v2 = ''; # partitioning attr. 1's value 2
my $pa2_id = 0; # attr. id for loop over @part_attr_2_list 
my $pa2_v = ''; # partitioning attr. 2's current value
my $pa12_v = ''; # joint part.attr. 1 & 2's current values
my $inst_ref = \[]; # pointer to instance

# Hash: Partitioning Attribute 1 - Goal Frequency, i.e.:
# goal frequency given each value of 1st partitioning attr.
# Hash_key: {$pa1_v}, Hash_value: 2-tuple: 
# [0]: pa1_v freq without goal, [1]: pa1_v freq with goal  
my %pa1_go_fr;

# Hash: Partitioning Attribute 1 - Goal Probability, i.e.:
# goal probability given each value of 1st partitioning attr.
# Hash_key: {$pa1_v}, Hash_value: goal prob given pa1_v, 
# computed from Hash %pa1_go_fr.
my %pa1_go_pr; 

# Hash: joint Partit. Attrs. 1 and 2 - Goal Frequency, i.e.:
# goal frequency given each pair of part.attr. 1 & 2's values
# Hash_key: {$pa1_v_$pa2_v}, Hash_value: 2-tuple: 
# [0]: pa1_v,pa2_v pair frequency without goal value
# [1]: pa1_v,pa2_v pair frequency with goal value
my %pa12_go_fr;

# Hash: joint Partit. Attrs. 1 and 2 - Goal Probability, i.e.:
# goal probab. given each pair of part.attr. 1 & 2's values
# Hash_key: {$pa1_v_$pa2_v}. Hash_value: goal prob. given 
# value pair: pa1_v,pa2_v. Computed from Hash %pa12_go_fr.
my %pa12_go_pr; 

# 1st Partit. Attr. effect on goal probability:
# more precisely: the effect, on goal value probability, of
# switching from first to second value of 1st Part. Attr.
# possible values: 'increase', 'decrease', or 'no_change'
my $pa1_eff_goal_prob = ''; 
# 2nd Partit. Attr. Value effect on goal probability
# more precisely: the effect, on goal value probability, of
# switching from first to second value of 1st Part. Attr.,
# given (conditioned on) each 2nd Part. Attr. value 
# possible values: 'increase', 'decrease', or 'no_change'
my @pa2val_eff_go_pr = (); # 1D array: [part.attr.2 value] 
# 1st Part. Attr.'s paradox occurrence count
my $pa1_parad_count = 0;

#############################################################
#  Print attribute and attribute value names in LOG file    #             
#############################################################
sub print_attr_val_names
{ print LOG "Attributes' names, types and values:\n";
  for (my $i=0; $i < $ATTR_COUNT; $i++)
  { print LOG "$i: $attr_names[$i], $attr_types[$i], ";
    if ($attr_types[$i] eq 'nominal')
    { print LOG "Values: @{$attr_values[$i]}\n";
    } elsif ( ($attr_types[$i] eq 'numeric')
              || ($attr_types[$i] eq 'string') )
    { print LOG "Values of type: $attr_values[$i]\n";
    } else
    { print LOG "ERROR: invalid attr. type: $attr_values[$i]",
                ", for attribute id: $i\n"; 
    }
  } # end of: 'for $i' (for all attributes)
}

#############################################################
# Print 1st partitioning attrib. effect on goal probability #
# (First, print goal probabilities given 1st partit. attr.) #
#############################################################
sub print_out_1stPA_eff_goal_prob
{ my ($ga, $pa1) = @_; # Goal Attr. id, Part. Attr. 1 id
  print OUTPUT 
        "Predictor Attribute: $pa1 ($attr_names[$pa1]), ",
        "Values: \{${$attr_values[$pa1]}[0] ",
        "${$attr_values[$pa1]}[1]\}\n";
  foreach $pa1_v (keys %pa1_go_pr)
  { printf OUTPUT "Label-of-Interest's probability given Predictor's value $pa1_v: %6.4f",
                  " $pa1_go_pr{$pa1_v}"; 
    print OUTPUT "\n";
  }
  print OUTPUT "When the Predictor's value changes from ",
    "${$attr_values[$pa1]}[0] to ${$attr_values[$pa1]}[1], ",
    "the Label-of-Interest's probability change is: ",
    "$pa1_eff_goal_prob\n"; 
}

#############################################################
# Print 2nd partitioning attrib. effect on goal probability #
# (First, print goal probab. given 1st and 2nd part. attrs.)#
#############################################################
sub print_out_2ndPA_eff_goal_prob
{ my ($pa1, $pa2) = @_; # Part. Attr. 1 id, Part. Attr. 2 id
  print OUTPUT 
        "Confounder Attribute: $pa2 ($attr_names[$pa2]),",
        " Values: \{${$attr_values[$pa2]}[0],", 
        " ${$attr_values[$pa2]}[1]\}\n";
  print OUTPUT "Label-of-Interest's probability for each ",
               "pair of Predictor's value and Confounder's ",
               "value (separated by '_'):\n";
  foreach $pa12_v (keys %pa12_go_pr)
  { printf OUTPUT "Label-of-Interest's probability given value pair $pa12_v: %6.4f",
          "$pa12_go_pr{$pa12_v}\n"; 
    print OUTPUT "\n";
  }
  foreach $pa2_v (@{$attr_values[$pa2]})
  { print OUTPUT "When the Confounder's value is $pa2_v, ",
     "and the Predictor's value changes from ",
     "${$attr_values[$pa1]}[0] to ${$attr_values[$pa1]}[1], ",
     "the Label-of-Interest's probability change is: ",
     "$pa2val_eff_go_pr[$pa2_v]\n"; 
  }
} 

##############################################################
# Read WEKA-format line with attribute name, type and values #
# and puts them in the corresponding arrays                  #
# Procedure called when ($expected_line_type eq 'attribute') #
# Assumes there is no space between attrib. values i"{...}"  #
##############################################################
sub read_WEKA_attr_line
{ if (/attribute\s+(\S+)\s+(\S+)/i)  
  { $attr_names[$attr_count] = $1; # $1 is attr. name in file
    $attr_values_or_type = $2;
  } else
  { print LOG "ERROR: Did not find attribute keyword\n";
  }
  # ASSUMES ATTRIBUTE IS NOMINAL, so $attr_values_or_type
  # contains a string of form "{value1, value2, ...}"
  if ($attr_values_or_type =~ (/\{(.+)\}/)) 
  { @{$attr_values[$attr_count]} = split /,\s*/, $1;
    $attr_types[$attr_count] = 'nominal';
  }
  elsif ($attr_values_or_type eq 'numeric')
  { $attr_values[$attr_count] = 'numeric_value';
    $attr_types[$attr_count] = 'numeric';
  }
  elsif ($attr_values_or_type eq 'string')
  { $attr_values[$attr_count] = 'string_value';
    $attr_types[$attr_count] = 'string';
  } else
  { print LOG "ERROR: Unknown type for attr.: $attr_count\n";
  }
}

##############################################################
# Load the dataset into 2D matrix: instances vs. attributes. #
##############################################################
sub read_instance
{ @instance = split /,\s*/; 
  if ( ($#instance + 1) != $ATTR_COUNT)
  { print LOG "ERROR:read ($#instance + 1) attributes,",
              " expected $ATTR_COUNT attributes\n";
  }
  push(@dataset, [@instance]);
}

################################
#        Main Program          #
################################    
# Read dataset from file and store it in array @dataset
# Previously initialized: $attr_count = 0; $inst_count = 0; 
while (<INPUT>)
{ chomp;
  unless ($_)
  { next;
  }
  if (/%/) # comment line
  { next;
  }
  if (/relation/i)
  { $expected_line_type = 'attribute';
    next;
  } elsif (/data/i)
  { $expected_line_type = 'instance';
    # all attribute names were read, save number of attributes
    $ATTR_COUNT = $attr_count;
    next;
  } elsif ($expected_line_type eq 'attribute')
  { &read_WEKA_attr_line();
    $attr_count++;
    next;
  } elsif ($expected_line_type eq 'instance') 
  { &read_instance();
    $inst_count++;
  }
} # end of: 'while <INPUT>'
&print_attr_val_names();
print OUTPUT
  "Goal Attribute: id: $GA_ID, Name: $attr_names[$GA_ID], ",
  "Values: @{$attr_values[$GA_ID]}, ",
  "Class label of Interest: Value: $GOAL_VAL, ",
  "Meaningful name: $GOAL_VAL_NAME\n",
  "List of Candidate Predictors (ids): @part_attr_1_list\n";

$ATTR_COUNT = $ATTR_COUNT - 1; # substracts class attribute
$INST_COUNT = $inst_count;
print LOG "$ATTR_COUNT predictor attributes read from file\n";
print LOG "$INST_COUNT instances read from input file.\n\n";

# 1st Partitioning Attribute effect on goal probability: 
# 'increase', 'decrease', 'no_change'
$pa1_eff_goal_prob = ''; 

### start of main code ###
if ($attr_types[$GA_ID] ne 'nominal')
{ print LOG "ERROR: goal attr. $GA_ID is not nominal\n";
  next;
}
foreach $pa1_id (@part_attr_1_list)
{ $pa1_parad_count = 0;
  if ($attr_types[$pa1_id] ne 'nominal')
  { print LOG "ERROR: Part.Att.1 $pa1_id is not nominal\n";
    next;
  }
  ### compute goal value frequency in set of instances
  ### where Part. Attr. 1's value is not missing.      
  foreach $inst_ref (@dataset) # dataset array: [inst][attr]
  { # update 1st Part. Attr. values' frequencies
    if (${$inst_ref}[$pa1_id] eq '?') # if missing value
    { print LOG "Program Failed: ERROR in input file: ",
            "Cannot cope with missing value flag ('?')\n";
      die "Cannot cope with missing value; program failed";
    }
    else # i.e., pa1_v is not a missing value
    { $pa1_v = ${$inst_ref}[$pa1_id];
      if (exists $pa1_go_fr{$pa1_v})
      { if (${$inst_ref}[$GA_ID] eq $GOAL_VAL)
        { $pa1_go_fr{$pa1_v}[1]++;
        } 
        else
        { $pa1_go_fr{$pa1_v}[0]++;
        }
      } # end of: 'if (hash key exists)'
      else # i.e., $pa1_go_fr{$pa1_v} does not exist
      { # initialize frequency of pa1_v without goal
        $pa1_go_fr{$pa1_v}[0] = 0; # increm. in next if
        # initialize frequency of pa1_v with goal
        $pa1_go_fr{$pa1_v}[1] = 0; # increm. in next if
        if (${$inst_ref}[$GA_ID] eq $GOAL_VAL)
        { $pa1_go_fr{$pa1_v}[1]++;
        } 
        else
        { $pa1_go_fr{$pa1_v}[0]++;
        }
      } # end of: 'else (i.e. hash key does not exist)'
    } # end of: 'else (i.e. pa1_v is not missing value)'
  } # end of: 'foreach $inst_ref'

  ### compute goal prob. for each value of 1st partit. attr.
  foreach $pa1_v (@{$attr_values[$pa1_id]})
  { if (exists $pa1_go_fr{$pa1_v}) 
    { if ($pa1_go_fr{$pa1_v}[1] != 0)
      { $pa1_go_pr{$pa1_v} = $pa1_go_fr{$pa1_v}[1] / 
        ($pa1_go_fr{$pa1_v}[0] + $pa1_go_fr{$pa1_v}[1]);
      } else
      { $pa1_go_pr{$pa1_v} = 0;
      }
    } # end of: 'if (hash key exists)'
    else
    { $pa1_go_pr{$pa1_v} = 0;
    } 
  } # end of: 'foreach $pa1_v'

  ### compute 1st partitioning attr. effect on goal probab.
  $pa1_eff_goal_prob = ''; 
  $pa1_v1 = ${$attr_values[$pa1_id]}[0]; 
  $pa1_v2 = ${$attr_values[$pa1_id]}[1]; 
  if ($pa1_go_pr{$pa1_v1} < $pa1_go_pr{$pa1_v2})
  { $pa1_eff_goal_prob = 'increase';
  }
  elsif ($pa1_go_pr{$pa1_v1} > $pa1_go_pr{$pa1_v2})
  { $pa1_eff_goal_prob = 'decrease';
  }
  else
  { $pa1_eff_goal_prob = 'no_change';
  }
  if ($pa1_eff_goal_prob eq 'no_change')
  { next; # jump to next iteration of: 'foreach $pa1_id'
  }  
  
  ### start of loop, for each partitioning attribute 2
  foreach $pa2_id (@part_attr_2_list)
  { if ($pa1_id == $pa2_id)
    { next; # ignore this $pa2 (the same as $pa1, no Paradox)
    }
    if ($attr_types[$pa2_id] ne 'nominal')
    { print LOG "ERROR: Part.Att.2 $pa2_id is not nominal\n";
      next;
    }
    ### compute goal value freq. in set of instances where
    ### both part.att. 1 value and pa2 value are not missing
    foreach $inst_ref (@dataset) # dataset array:[inst][att]
    { if (${$inst_ref}[$pa1_id] eq '?') # if missing value
      { print LOG "Program Failed: ERROR in input file: ",
            "Cannot cope with missing value flag ('?')\n";
        die "Cannot cope with missing value, program failed"; 
      }
      if (${$inst_ref}[$pa2_id] eq '?') # if missing value
      { print LOG "Program Failed: ERROR in input file: ",
            "Cannot cope with missing value flag ('?')\n";
        die "Cannot cope with missing value, program failed";
      }
      $pa1_v = ${$inst_ref}[$pa1_id];
      $pa2_v = ${$inst_ref}[$pa2_id];
      $pa12_v = $pa1_v . "_" . $pa2_v; #joint hash key
      if (exists $pa12_go_fr{$pa12_v})
      { if (${$inst_ref}[$GA_ID] eq $GOAL_VAL)
        { $pa12_go_fr{$pa12_v}[1]++;
        } 
        else
        { $pa12_go_fr{$pa12_v}[0]++;
        }
      } # end of: 'if (hash key exists)'
      else # i.e., $pa12_go_fr{$pa12_v} does not exist
      { # initialize frequency of pa12_v without goal
        $pa12_go_fr{$pa12_v}[0] = 0; # increm. in next if
        # initialize frequency of pa12_v with goal
        $pa12_go_fr{$pa12_v}[1] = 0; # increm. in next if
        if (${$inst_ref}[$GA_ID] eq $GOAL_VAL)
        { $pa12_go_fr{$pa12_v}[1]++;
        } 
        else
        { $pa12_go_fr{$pa12_v}[0]++;
        }
      } # end of: 'else (i.e. hash key does not exist)'
    } # end of: 'foreach $inst_ref' 
 
    ### compute goal prob. for each pair of part.attr. vals
    foreach $pa1_v (@{$attr_values[$pa1_id]})
    { foreach $pa2_v (@{$attr_values[$pa2_id]})
      { $pa12_v = $pa1_v . "_" . $pa2_v;
        if (exists $pa12_go_fr{$pa12_v}) 
        { if ($pa12_go_fr{$pa12_v}[1] != 0)
          { $pa12_go_pr{$pa12_v} = $pa12_go_fr{$pa12_v}[1] 
               / ( $pa12_go_fr{$pa12_v}[0] 
                 + $pa12_go_fr{$pa12_v}[1] );
          } else
          { $pa12_go_pr{$pa12_v} = 0;
          }
        } # end of: 'if (hash key exists for $pa12_go_fr)'
        else # i.e., $pa12_v does not occur in input file
        { $pa12_go_fr{$pa12_v}[0] = 0; 
          $pa12_go_fr{$pa12_v}[1] = 0;
          $pa12_go_pr{$pa12_v} = 0;
        } 
      } # end of: 'foreach $pa2_v'
    } # end of: 'foreach $pa1_v'

    # foreach part-attr-2 value, compare goal probabil. in
    # sub-pop with part-att-2 value & part-attr-1-val-1
    # vs sub-pop with part-att-2 value & part-attr-1-val-2.
    # If the observed effect on goal probabil. was opposite
    # to the effect in pa1_eff_goal_prob for all values of
    # part-attr-2, then the Paradox has occurred.
      
    my $paradox_found = 'true'; # default value, can change
      
    my $pa12_v1; # joint hash key: pa1 value1 and pa2 value
    my $pa12_v2; # joint hash key: pa1 value2 and pa2 value
    # pa2 value effect on gp: increase/decrease/no_change
 
    foreach $pa2_v (@{$attr_values[$pa2_id]})
    { $pa1_v1 = ${$attr_values[$pa1_id]}[0];
      $pa1_v2 = ${$attr_values[$pa1_id]}[1]; 
      $pa12_v1 = $pa1_v1 . "_" . $pa2_v; # joint key
      $pa12_v2 = $pa1_v2 . "_" . $pa2_v; # joint key
      $pa2val_eff_go_pr[$pa2_v] = '';
      if (! exists $pa12_go_pr{$pa12_v1}) 
      { print LOG "ERROR: pa12_v1 key not in hash\n";
      } 
      elsif (! exists $pa12_go_pr{$pa12_v2})
      { print LOG "ERROR: pa12_v12 key not in hash\n";
      }
      elsif ($pa12_go_pr{$pa12_v1} < $pa12_go_pr{$pa12_v2})
      { $pa2val_eff_go_pr[$pa2_v] = 'increase';
      }
      elsif ($pa12_go_pr{$pa12_v1} > $pa12_go_pr{$pa12_v2})
      { $pa2val_eff_go_pr[$pa2_v] = 'decrease';
      }
      else
      { $pa2val_eff_go_pr[$pa2_v] = 'no_change';
      }
      ### checked before: $pa1_eff_goal_prob ne 'no_change'
      if ( ($pa1_eff_goal_prob eq $pa2val_eff_go_pr[$pa2_v])
           || ($pa2val_eff_go_pr[$pa2_v] eq 'no_change') )
      { $paradox_found = 'false';
        last; # abandon loop: 'foreach $pa2_v'
      }
    } # end of: 'foreach $pa2_v'
    if ($paradox_found eq 'true')
    { $pa1_parad_count++;
      print OUTPUT "\nPARADOX FOUND: ",
      "Predictor Attribute: $pa1_id ($attr_names[$pa1_id]), ",
      "Confounder Attribute: $pa2_id ($attr_names[$pa2_id])\n",
      "Goal Attribute: $GA_ID ($attr_names[$GA_ID]); ",
      "Class label of Interest: $GOAL_VAL ",
      "($GOAL_VAL_NAME)\n=============\n";
      &print_out_1stPA_eff_goal_prob($GA_ID,$pa1_id);
      &print_out_2ndPA_eff_goal_prob($pa1_id,$pa2_id);
    } # end of: if ($paradox_found eq 'true')
 
    %pa12_go_fr = (); 
    %pa12_go_pr = ();
  } # end of: 'foreach $pa2_id'
  
  if ($pa1_parad_count)
  { print OUTPUT "\nAttr $pa1_id ($attr_names[$pa1_id]) is ",
         "the Predictor Attribute in $pa1_parad_count ";
    if ($pa1_parad_count > 1)
    { print OUTPUT "paradoxes\n";
    } 
    else
    { print OUTPUT "paradox\n";
    }
  }

  %pa1_go_fr = (); 
  %pa1_go_pr = ();
} # end of: 'foreach $pa1_id'

close INPUT;
close OUTPUT;
close LOG;

